"""
Classes and functions for analyzing and visualizing the relations of two panel categories.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Union, Tuple
from scipy import stats
import statsmodels.api as sm
import warnings

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import categories_df
from macrosynergy.management.utils import apply_slip as apply_slip_util
from macrosynergy.management.types import QuantamentalDataFrame


class CategoryRelations(object):
    """
    Class for analyzing and visualizing the relations of multiple panel categories.

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized DataFrame with the necessary columns: 'cid', 'xcat', 'real_date'
        and at least one column with values of interest.
    xcats : List[str]
        exactly two extended categories to be analyzed. If there is a hypothesized
        explanatory-dependent relation, the first category is the explanatory variable and
        the second category the explained variable.
    cids : List[str]
        cross-sections for which the category relations is being analyzed. Default is
        all in the DataFrame.
    start : str
        earliest date in ISO format. Default is None in which case the earliest date in
        the DataFrame will be used.
    end : str
        latest date in ISO format. Default is None in which case the latest date in the
        DataFrame will be used.
    blacklist : dict
        cross-sections with date ranges that should be excluded from the analysis.
    years : int
        number of years over which data are aggregated. Supersedes the 'freq' parameter
        and does not allow lags, Default is None, meaning no multi-year aggregation. Note:
        for single year labelled plots, better use freq='A' for cleaner labels.
    val : str
        name of column that contains the values of interest. Default is 'value'.
    freq : str
        letter denoting frequency at which the series are to be sampled. This must be
        one of 'D', 'W', 'M', 'Q', 'A'. Default is 'M'.
    lag : int
        lag (delay of arrival) of first (explanatory) category in periods as set by
        freq. Default is 0. Importantly, for analyses with explanatory and dependent
        categories, the first category takes the role of the explanatory and a positive lag
        means that the explanatory values will be deferred into the future, i.e. relate to
        future values of the explained variable.
    xcat_aggs : List[str]
        Exactly two aggregation methods. Default is 'mean' for both.
    xcat1_chg : str
        time series changes are applied to the first category. Default is None. Options
        are 'diff' (first difference) and 'pch' (percentage change). The changes are
        calculated over the number of periods determined by `n_periods`.
    n_periods : int
        number of periods over which changes of the first category have been calculated.
        Default is 1.
    fwin : int
        forward moving average window of second category. Default is 1, i.e no average.
        Importantly, for analysis with explanatory and dependent categories, the second
        takes the role of the dependent and a forward window means that the dependent values
        average forward into the future.
    xcat_trims : List[float]
        two-element list with maximum absolute values for the two respective categories.
        Observations with higher values will be trimmed, i.e. removed from the analysis (not
        winsorized!). Default is None for both. Trimming is applied after all other
        transformations.
    slip : int
        number of periods to 'slip' the explanatory variable, i.e. the first category.
        Here, slip mimics the late arrival of the data, or the time it takes
        to act on the data. Default is 0. 
    """

    def __init__(
        self,
        df: pd.DataFrame,
        xcats: List[str],
        cids: List[str] = None,
        val: str = "value",
        start: str = None,
        end: str = None,
        blacklist: dict = None,
        years: int = None,
        freq: str = "M",
        lag: int = 0,
        fwin: int = 1,
        xcat_aggs: List[str] = ["mean", "mean"],
        xcat1_chg: str = None,
        n_periods: int = 1,
        xcat_trims: List[float] = [None, None],
        slip: int = 0,
    ):
        """Initializes CategoryRelations"""

        if not isinstance(freq, str):
            raise TypeError("freq must be a string.")

        self.xcats: List[str] = xcats
        self.cids: List[str] = cids
        self.val: str = val
        self.freq: str = freq.upper()
        self.lag: int = lag
        self.years: int = years
        self.aggs: List[str] = xcat_aggs
        self.xcat1_chg: str = xcat1_chg
        self.n_periods: int = n_periods
        self.xcat_trims: List[float] = xcat_trims
        self.slip: int = slip

        if self.freq not in ["D", "W", "M", "Q", "A"]:
            raise ValueError("freq must be one of 'D', 'W', 'M', 'Q', 'A'.")
        if not isinstance(val, str):
            raise TypeError("val must be a string.")
        if not {"cid", "xcat", "real_date", val}.issubset(set(df.columns)):
            raise ValueError(
                f"`df` must have columns 'cid', 'xcat', 'real_date' and `{val}`."
            )
        df = QuantamentalDataFrame(df)

        if not isinstance(xcats, (list, tuple)):
            raise TypeError("`xcats` must be a list or a tuple.")
        elif not len(xcats) == 2:
            raise ValueError("`xcats` must have exactly two elements.")
        if not isinstance(slip, int):
            raise TypeError("`slip` must be a non-negative integer.")
        elif slip < 0:
            raise ValueError("`slip` must be a non-negative integer.")

        if not isinstance(xcat_aggs, (list, tuple)):
            raise TypeError("xcat_aggs must be a list or a tuple.")

        # copy DF to avoid side-effects
        df: pd.DataFrame = df.copy()
        # Select the cross-sections available for both categories.
        df.loc[:, "real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")

        if self.slip != 0:
            metrics_found: List[str] = list(
                set(df.columns) - set(["cid", "xcat", "real_date"])
            )
            # here, the slip is applied to the the first xcat (explanatory variable)
            df = self.apply_slip(
                df=df,
                slip=self.slip,
                cids=self.cids,
                xcats=[self.xcats[0]],
                metrics=metrics_found,
            )

        # capture warning from intersection_cids, in case the two categories do not
        # share any cross-sections.
        warnings_list = []
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            shared_cids = CategoryRelations.intersection_cids(df, xcats, cids)
            for warning in w:
                warnings_list.append(str(warning.message))

        # if shared_cids is empty, then the analysis is not possible.
        # The warning from intersection_cids now becomes an error.
        if len(shared_cids) == 0:
            error_message = "The two categories have no shared cross-sections."
            if len(warnings_list) > 0:
                error_message += f"\nPossible reason(s) for error: "
                error_message += "\n".join(warnings_list)

            error_message += "\nPlease check input parameters."
            raise ValueError(error_message)

        # Will potentially contain NaN values if the two categories are defined over
        # time-periods.
        df = categories_df(
            df,
            xcats,
            shared_cids,
            val=val,
            start=start,
            end=end,
            freq=self.freq,
            blacklist=blacklist,
            years=years,
            lag=lag,
            fwin=fwin,
            xcat_aggs=xcat_aggs,
        )

        if xcat1_chg is not None:
            xcat1_error = (
                "Change applied to the explanatory variable must either be "
                "first-differencing, 'diff', or percentage change, 'pch'."
            )
            assert xcat1_chg in ["diff", "pch"], xcat1_error
            n_periods_error = f"<int> expected and not {type(n_periods)}."
            assert isinstance(n_periods, int), n_periods_error

            df = CategoryRelations.time_series(
                df,
                change=xcat1_chg,
                n_periods=n_periods,
                shared_cids=shared_cids,
                expln_var=xcats[0],
            )

        if any([xt is not None for xt in self.xcat_trims]):
            xcat_trim_error = (
                "Two values expected corresponding to the number " "of categories."
            )
            assert len(xcat_trims) == len(xcats), xcat_trim_error

            types = [
                isinstance(elem, (float, int)) and elem >= 0.0 for elem in xcat_trims
            ]
            assert any(types), "Expected two floating point values."

            df = CategoryRelations.outlier_trim(df, xcats, xcat_trims)

        # NaN values will not be handled if both of the above conditions are not
        # satisfied.
        self.df = df.dropna(axis=0, how="any")

    @classmethod
    def intersection_cids(cls, df, xcats, cids):
        """
        Returns common cross-sections across both categories and specified parameter.

        Parameters
        ----------
        df : ~pandas.DataFrame
            standardised DataFrame.
        xcats : List[str]
            exactly two extended categories to be checked on.
        cids : List[str]
            cross-sections for which the category relation is being analyzed.

        Returns
        -------
        List[str]
            usable: List of the common cross-sections across the two categories.
        """

        set_1 = set(df[df["xcat"] == xcats[0]]["cid"])
        set_2 = set(df[df["xcat"] == xcats[1]]["cid"])

        miss_1 = list(set(cids).difference(set_1))
        miss_2 = list(set(cids).difference(set_2))

        if len(miss_1) > 0:
            print(f"{xcats[0]} misses: {sorted(miss_1)}.")
            warnings.warn(f"{xcats[0]} misses: {sorted(miss_1)}.", UserWarning)
        if len(miss_2) > 0:
            print(f"{xcats[1]} misses: {sorted(miss_2)}.")
            warnings.warn(f"{xcats[1]} misses: {sorted(miss_2)}.", UserWarning)

        usable = list(set_1.intersection(set_2).intersection(set(cids)))

        return usable

    @staticmethod
    def apply_slip(
        df: pd.DataFrame,
        slip: int,
        cids: List[str],
        xcats: List[str],
        metrics: List[str],
    ) -> pd.DataFrame:
        """
        Calls the utility function apply_slip_util defined in df_utils.
        """
        return apply_slip_util(
            df=df, slip=slip, cids=cids, xcats=xcats, metrics=metrics, raise_error=False
        )

    @classmethod
    def time_series(
        cls,
        df: pd.DataFrame,
        change: str,
        n_periods: int,
        shared_cids: List[str],
        expln_var: str,
    ):
        """
        Apply time-series changes to the explanatory variable. Calculates first
        differences or percentage changes of the time series.

        Parameters
        ----------
        df : ~pandas.DataFrame
            multi-indexed DataFrame hosting the two categories: first column represents
            the explanatory variable; second column hosts the dependent variable. The
            DataFrame's index is the real-date and cross-section.
        change : str
            type of change to be applied. Can be 'diff' for first-differencing or 'pch'
            for percentage change.
        n_periods : int
            number of base periods in df over which the change is applied.
        shared_cids : List[str]
            shared cross-sections across the two categories and the received list.
        expln_var : str
            only the explanatory variable's data series will be changed from the raw
            value series to a difference or percentage change value.

        Returns
        -------
        ~pandas.DataFrame
            returns the same multi-indexed DataFrame but with an adjusted series
            inline with the 'change' parameter.
        """

        if change not in ["diff", "pch"]:
            raise ValueError("change must be 'diff' or 'pch'.")

        df_lists = []
        for c in shared_cids:
            temp_df: pd.DataFrame = df.loc[c].copy()

            if change == "diff":
                temp_df[expln_var] = temp_df[expln_var].diff(periods=n_periods)
            elif change == "pch":
                temp_df[expln_var] = temp_df[expln_var].pct_change(periods=n_periods)

            temp_df["cid"] = c
            temp_df = temp_df.set_index("cid", append=True)
            df_lists.append(temp_df)

        df_ = pd.concat(df_lists)
        df_ = df_.dropna(axis=0, how="any")
        return df_

    @classmethod
    def outlier_trim(cls, df: pd.DataFrame, xcats: List[str], xcat_trims: List[float]):
        """
        Trim outliers from the dataset.

        Parameters
        ----------
        df : ~pandas.DataFrame
            multi-indexed DataFrame hosting the two categories. The transformations, to
            each series, have already been applied.
        xcats : List[str]
            explanatory and dependent variable.
        xcat_trims : List[float]
            two-element list with maximum absolute values for the two respective
            categories. Observations with higher values will be trimmed, i.e. removed from
            the analysis (not winsorized!).


        Returns
        -------
        ~pandas.DataFrame
            returns the same multi-indexed DataFrame.


        .. note::
            Outliers are classified as any datapoint whose absolute value exceeds the
            predefined value specified in the field self.xcat_trims. The values will be
            set to NaN, and subsequently excluded from any regression modelling or
            correlation coefficients.
        """

        xcat_dict = dict(zip(xcats, xcat_trims))

        for k, v in xcat_dict.items():
            # if the trim value is None, then leave the series as is
            if v is not None:
                df[k] = np.where(np.abs(df[k]) < v, df[k], np.nan)

        df = df.dropna(axis=0, how="any")
        return df

    def corr_prob_calc(
        self, df_probability: Union[pd.DataFrame, List[pd.DataFrame]], prob_est: str
    ):
        """
        Compute the correlation coefficient and probability statistics.

        Parameters
        ----------
        df_probability : List[~pandas.DataFrame] or ~pandas.DataFrame
            pandas DataFrame containing the dependent and explanatory variables.
        prob_est : str
            type of estimator for probability of significant relation.

        Returns
        -------
        List[tuple(float, float)]


        .. note::
            The method is able to handle multiple DataFrames, and will return the
            corresponding number of statistics held inside a List.
        """

        if isinstance(df_probability, pd.DataFrame):
            df_probability = [df_probability]

        cpl = []
        for i, df_i in enumerate(df_probability):
            feat = df_i[self.xcats[0]].to_numpy()
            targ = df_i[self.xcats[1]].to_numpy()
            coeff, pval = stats.pearsonr(feat, targ)
            if prob_est == "kendall":
                _, pval = stats.kendalltau(feat, targ)
            if prob_est == "map":
                X = df_i.loc[:, self.xcats[0]]
                X = sm.add_constant(X)
                y = df_i.loc[:, self.xcats[1]]
                groups = df_i.reset_index().real_date
                re = sm.MixedLM(y, X, groups).fit(reml=False)  # random effects est
                pval = float(re.summary().tables[1].iloc[1, 3])
            row = [np.round(coeff, 3), np.round(1 - pval, 3)]
            cpl.append(row)
        return cpl

    def corr_probability(
        self,
        df_probability: Union[pd.DataFrame, List[pd.DataFrame]],
        prob_est: str,
        time_period: str = "",
        coef_box_loc: str = "upper left",
        ax: plt.Axes = None,
    ):
        """
        Add the computed correlation coefficient and probability to a Matplotlib table.

        Parameters
        ----------
        df_probability : List[~pandas.DataFrame] or ~pandas.DataFrame
            pandas DataFrame containing the dependent and explanatory variables. Able to
            handle multiple DataFrames representing different time-periods of the original
            series.
        prob_est : str
            type of estimator for probability of significant relation.
        time_period : str
            indicator used to clarify which time-period the statistics are computed for.
            For example, before 2010 and after 2010: the two periods experience very
            different macroeconomic conditions. The default is an empty string.
        coef_box_loc : str
            location on the graph of the aforementioned box. The default is in the upper
            left corner.
        prob_bool : bool
            boolean parameter which determines whether the probability value is included
            in the table. The default is True.
        ax : plt.Axes
            Matplotlib Axes object. If None (default), new axes will be created.
        """

        time_period_error = f"<str> expected - received {type(time_period)}."
        assert isinstance(time_period, str), time_period_error

        cpl = self.corr_prob_calc(df_probability=df_probability, prob_est=prob_est)

        fields = [
            f"Correlation\n coefficient {time_period}",
            f"Probability\n of significance {time_period}",
        ]

        if isinstance(df_probability, list) and len(df_probability) == 2:
            row_headers = ["Before 2010", "After 2010"]
            cellC = [
                ["lightsteelblue", "lightsteelblue"],
                ["lightsalmon", "lightsalmon"],
            ]
        else:
            row_headers = None
            cellC = None

        if ax is None:
            data_table = plt.table(
                cellText=cpl,
                cellColours=cellC,
                colLabels=fields,
                cellLoc="center",
                loc=coef_box_loc,
                zorder=10,
            )
        else:
            data_table = ax.table(
                cellText=cpl,
                cellColours=cellC,
                colLabels=fields,
                cellLoc="center",
                loc=coef_box_loc,
                zorder=10,
            )

        return data_table

    def annotate_facet(self, data, **kws):
        """Annotate each graph within the facet grid."""

        x = data[self.xcats[0]].to_numpy()
        y = data[self.xcats[1]].to_numpy()
        coeff, pval = stats.pearsonr(x, y)

        cpl = np.round(coeff, 3)
        fields = "Correlation coefficient: "
        ax = plt.gca()
        ax.text(0.04, 0.1, f"{fields} {cpl}", fontsize=10, transform=ax.transAxes)

    def reg_scatter(
        self,
        title: str = None,
        title_fontsize: int = 14,
        labels: bool = False,
        size: Tuple[float] = None,
        xlab: str = None,
        ylab: str = None,
        label_fontsize: int = 12,
        tick_fontsize: int = 12,
        coef_box: str = None,
        coef_box_size: Tuple[float] = (0.4, 2.5),
        coef_box_font_size: int = 0,
        prob_est: str = "pool",
        fit_reg: bool = True,
        reg_ci: int = 95,
        reg_order: int = 1,
        reg_robust: bool = False,
        separator: Union[str, int] = None,
        title_adj: float = 1,
        single_chart: bool = False,
        single_scatter: bool = False,
        ncol: int = None,
        ax: plt.Axes = None,
        remove_zero_predictor: bool = False,
    ):
        """
        Display scatter-plot and regression line.

        Parameters
        ----------
        title : str
            title of plot. If None (default) an informative title is applied.
        labels : bool
            assign a cross-section/period label to each dot. Default is False.
        size : Tuple[float]
            width and height of the figure
        xlab : str
            x-axis label. Default is no label.
        ylab : str
            y-axis label. Default is no label.
        fit_reg : bool
            if True (default) adds a regression line.
        reg_ci : int
            size of the confidence interval for the regression estimate. Default is 95.
            Can be None.
        reg_order : int
            order of the regression equation. Default is 1 (linear).
        reg_robust : bool
            if this will de-weight outliers, which is computationally expensive. Default
            is False.
        coef_box : str
            two-purpose parameter. Firstly, if the parameter equals None, the
            correlation coefficient and probability statistics will not be included in the
            graphic. Secondly, if the statistics are to be included, pass in the desired
            location on the graph which, in addition, will act as a pseudo-boolean
            parameter. The options are standard, i.e. 'upper left', 'lower right' and so
            forth. Default is None, i.e the statistics are not displayed.
        prob_est : str
            type of estimator for probability of significant relation.

            - "pool" (default)
                which means that all observation are treated as
                independent and calculates Pearson's correlation coefficient.
            - "map"
                denoting Macrosynergy panel test. This is based on a panel regression
                with period-specific random effects and greatly mitigates the issue of
                pseudo-replication if panel features and targets are correlated across
                time.
                See also https://research.macrosynergy.com/testing-macro-trading-factors/
            - "kendall"
                which calculates the Kendall rank correlation coefficient. It is
                a non-parametric statistic used to measure the strength and direction of
                association between two ranked variables.
        separator : Union[str, int]
            allows categorizing the scatter analysis by cross-section or integer. In the
            former case the argument is set to "cids" and in the latter case the argument is
            set to a year [2010, for instance] which will subsequently split the time-period
            into the sample before (not including) that year and from (including) that year.
        title_adj : float
            parameter that sets top of figure to accommodate title. Default is 1.
        single_chart : bool
            boolean parameter determining whether the x- and y- labels are only written
            on a single graph of the Facet Grid (useful if there are numerous charts, and
            the labels are excessively long). The default is False, and the names of the
            axis will be displayed on each grid if not conflicting with the label for each
            variable.
        ncol : int
            number of columns in the facet grid. Default is None, in which case the
            number of columns is determined by the number of cross-sections.
        ax : plt.Axes
            Matplotlib Axes object. If None (default), new figure and axes objects will
            be created. If an Axes object is passed, the plot will be drawn on the Axes, and
            plt.show() will not be called.
        remove_zero_predictor : bool, default=False
            Remove zeros from the input series.
        """

        coef_box_loc_error = (
            "The parameter expects a string used to delimit the "
            "location of the box: 'upper left', 'lower right' etc."
        )
        if coef_box is not None:
            assert isinstance(coef_box, str), coef_box_loc_error

        assert prob_est in [
            "pool",
            "map",
            "kendall",
        ], "prob_est must be 'pool', 'kendall' or 'map'"

        if len(self.cids) == 1 and prob_est == "map":
            warnings.warn(
                "The 'map' estimator is not applicable to a single cross-section. "
                "Using 'pool' instead.",
                UserWarning,
            )
            prob_est = "pool"

        sns.set_theme(style="whitegrid")
        dfx = self.df.copy()

        if title is None and (self.years is None):
            dates = (
                self.df.index.get_level_values("real_date")
                .to_series()
                .dt.strftime("%Y-%m-%d")
            )
            title = (
                f"{self.xcats[0]} and {self.xcats[1]} "
                f"from {dates.min()} to {dates.max()}"
            )
        elif title is None:
            title = f"{self.xcats[0]} and {self.xcats[1]}"

        if ax is not None:
            if not isinstance(ax, plt.Axes):
                raise TypeError("ax must be a matplotlib Axes object.")
            show_plot = False
        else:
            show_plot = True

        set_font_size = False
        if not (isinstance(coef_box_font_size, int) and coef_box_font_size >= 0):
            raise ValueError("coef_box_font_size must be a non-negative integer.")
        if coef_box_font_size == 0:
            set_font_size = True
            coef_box_font_size = 12

        # If "separator" is type Integer, the scatter plot is split across two
        # time-periods where the divisor is the received year.
        if size is None:
            size = (3, 3) if separator == "cids" else (12, 8)
        else:
            if (
                not isinstance(size, tuple)
                or len(size) != 2
                or not all(isinstance(i, (int, float)) for i in size)
            ):
                raise TypeError("size must be a tuple of ints/floats.")

        if isinstance(separator, int):
            year_error = "Separation by years does not work with year groups."
            assert self.years is None, year_error

            if ax is None:
                fig, ax = plt.subplots(figsize=size)

            if "real_date" not in dfx.index.names:
                raise ValueError("`real_date` expected in index names.")

            if remove_zero_predictor:
                dfx = dfx[dfx.loc[:, self.xcats[0]] != 0]

            rdt_index = list(dfx.index.names).index("real_date")
            index_years = dfx.index.get_level_values(rdt_index).year
            years_in_df = list(index_years.unique())

            assert separator in years_in_df, "Separator year is not in the range."
            error_sep = "Separator year must not be the first in the range."
            assert separator > np.min(years_in_df), error_sep

            label_set1 = f"before {separator}"
            label_set2 = f"from {separator}"
            dfx1 = dfx[index_years < separator]
            dfx2 = dfx[index_years >= separator]

            sns.regplot(
                data=dfx1,
                x=self.xcats[0],
                y=self.xcats[1],
                ci=reg_ci,
                order=reg_order,
                robust=reg_robust,
                fit_reg=fit_reg,
                scatter_kws={"s": 30, "alpha": 0.5},
                label=label_set1,
                line_kws={"lw": 1},
                ax=ax,
            )
            sns.regplot(
                data=dfx2,
                x=self.xcats[0],
                y=self.xcats[1],
                ci=reg_ci,
                order=reg_order,
                robust=reg_robust,
                fit_reg=fit_reg,
                label=label_set2,
                scatter_kws={"s": 30, "alpha": 0.5},
                line_kws={"lw": 1},
                ax=ax,
            )

            if coef_box is not None:
                data_table = self.corr_probability(
                    df_probability=[dfx1, dfx2],
                    time_period="",
                    coef_box_loc=coef_box,
                    prob_est=prob_est,
                    ax=ax,
                )
                x_scale = coef_box_size[0]
                y_scale = coef_box_size[1]
                data_table.scale(x_scale, y_scale)
                data_table.auto_set_font_size(set_font_size)
                data_table.set_fontsize(coef_box_font_size)

            ax.legend(loc="upper right")
            ax.set_title(title, fontsize=title_fontsize)
            if xlab is not None:
                ax.set_xlabel(xlab, fontsize=label_fontsize)
            if ylab is not None:
                ax.set_ylabel(ylab, fontsize=label_fontsize)

        elif separator == "cids" and not single_scatter:
            assert isinstance(single_chart, bool)

            dfx_copy = dfx.reset_index().rename(columns={"level_0": "cid"})
            n_cids = len(dfx_copy["cid"].unique())

            error_cids = (
                "There must be more than one cross-section to use "
                "separator = 'cids'."
            )
            assert n_cids > 1, error_cids

            # "Wrap" the column variable at this width, so that the column facets span
            # multiple rows. Used to determine the number of grids on each row.
            dict_coln = {2: 2, 5: 3, 8: 4, 30: 5}

            keys_ar = np.array(list(dict_coln.keys()))
            key = keys_ar[keys_ar <= n_cids][-1]
            if ncol is None:
                ncol = dict_coln[key]
            if ncol > n_cids:
                ncol = n_cids

            # The DataFrame is already a standardised DataFrame. Three columns: two
            # categories (dependent & explanatory variable) and the respective
            # cross-sections. The index will be the date timestamp.

            facet_height = size[1]  # height of each facet in inches
            facet_aspect = size[0] / size[1]  # aspect ratio of each facet

            fg = sns.FacetGrid(
                data=dfx_copy,
                col="cid",
                col_wrap=ncol,
                height=facet_height,
                aspect=facet_aspect,
            )
            fg.map(
                sns.regplot,
                self.xcats[0],
                self.xcats[1],
                ci=reg_ci,
                order=reg_order,
                robust=reg_robust,
                fit_reg=fit_reg,
                scatter_kws={"s": 15, "alpha": 0.5, "color": "lightgray"},
                line_kws={"lw": 1},
            )

            if coef_box is not None:
                fg.map_dataframe(self.annotate_facet)

            fg.set_titles(col_template="{col_name}")
            fg.fig.suptitle(title, y=title_adj, fontsize=14)

            if not single_chart:
                if xlab is not None:
                    fg.set_xlabels(xlab, clear_inner=True)
                if ylab is not None:
                    fg.set_ylabels(ylab)
            else:
                error = "Label expected for the respective axis."
                assert xlab is not None, error
                assert ylab is not None, error
                number_of_graphs = len(fg.axes)
                no_columns = fg._ncol
                remainder = int(number_of_graphs % no_columns)

                for i in range(number_of_graphs):
                    fg.axes[i].set_xlabel("")
                    fg.axes[i].set_ylabel("")

                    if remainder == 0:
                        fg.axes[no_columns - 1].set_xlabel(xlab)
                        fg.axes[no_columns - 1].set_ylabel(ylab)
                    else:
                        fg.axes[-remainder].set_xlabel(xlab)
                        fg.axes[-remainder].set_ylabel(ylab)

        elif separator == "cids" and single_scatter:
            assert isinstance(single_chart, bool)

            if (
                coef_box == "upper right"
            ):  # Since otherwise this overlaps with cid legend
                coef_box = "upper left"

            dfx_copy = dfx.reset_index()
            cids = dfx_copy["cid"].unique()
            n_cids = len(cids)

            error_cids = (
                "There must be more than one cross-section to use "
                "separator = 'cids'."
            )
            assert n_cids > 1, error_cids

            if ax is None:
                fig, ax = plt.subplots(figsize=size)

            # Perform a single global regression
            sns.regplot(
                data=dfx_copy,
                x=self.xcats[0],
                y=self.xcats[1],
                ci=reg_ci,
                order=reg_order,
                robust=reg_robust,
                fit_reg=fit_reg,
                scatter=False,  # Do not plot scatter points in regplot
                line_kws={"lw": 1, "color": "black"},
                ax=ax,
            )

            # Color code the scatter points by cid
            for i, cid in enumerate(cids):
                dfx_i = dfx_copy[dfx_copy["cid"] == cid]
                ax.scatter(
                    dfx_i[self.xcats[0]],
                    dfx_i[self.xcats[1]],
                    label=f"{cid}",
                    s=30,
                    alpha=0.5,
                )

            if coef_box is not None:
                data_table = self.corr_probability(
                    df_probability=dfx_copy,
                    time_period="",
                    coef_box_loc=coef_box,
                    prob_est=prob_est,
                    ax=ax,
                )
                x_scale = coef_box_size[0]
                y_scale = coef_box_size[1]
                data_table.scale(x_scale, y_scale)
                data_table.auto_set_font_size(set_font_size)
                data_table.set_fontsize(coef_box_font_size)

            ax.legend(loc="upper right", title="Cids")
            ax.set_title(title, fontsize=title_fontsize)
            if xlab is not None:
                ax.set_xlabel(xlab, fontsize=label_fontsize)
            if ylab is not None:
                ax.set_ylabel(ylab, fontsize=label_fontsize)

        elif separator is None:
            if ax is None:
                fig, ax = plt.subplots(figsize=size)
            else:
                show_plot = False

            if remove_zero_predictor:
                dfx = dfx[dfx.loc[:, self.xcats[0]] != 0]

            sns.regplot(
                data=dfx,
                x=self.xcats[0],
                y=self.xcats[1],
                ci=reg_ci,
                order=reg_order,
                robust=reg_robust,
                fit_reg=fit_reg,
                scatter_kws={"s": 30, "alpha": 0.5, "color": "lightgray"},
                line_kws={"lw": 1},
                ax=ax,
            )

            if coef_box is not None:
                data_table = self.corr_probability(
                    df_probability=dfx,
                    prob_est=prob_est,
                    coef_box_loc=coef_box,
                    ax=ax,
                )
                x_scale = coef_box_size[0]
                y_scale = coef_box_size[1]
                data_table.scale(x_scale, y_scale)
                data_table.auto_set_font_size(set_font_size)
                data_table.set_fontsize(coef_box_font_size)

            if labels:
                error_freq = "Labels only available for monthly or lower frequencies."
                assert self.freq in ["A", "Q", "M"], error_freq

                df_labs = dfx.dropna().index.to_frame(index=False)
                if "cid" not in df_labs.columns:
                    df_labs = df_labs.rename(columns={0: "cid"})
                if self.years is not None:
                    ser_labs = (
                        df_labs["cid"].astype("object") + " " + df_labs["real_date"]
                    )
                else:
                    ser_labs = df_labs["cid"].astype("object") + " "
                    ser_labs += df_labs["real_date"].dt.year.astype("string")
                    if self.freq == "Q":
                        ser_labs += "Q" + df_labs["real_date"].dt.quarter.astype(
                            "string"
                        )

                    elif self.freq == "M":
                        ser_labs += "-" + df_labs["real_date"].dt.month.astype("string")

                for i in range(dfx.shape[0]):
                    ax.text(
                        x=dfx[self.xcats[0]][i] + 0,
                        y=dfx[self.xcats[1]][i] + 0,
                        s=ser_labs[i],
                        fontdict=dict(color="black", size=8),
                    )

            ax.set_title(title, fontsize=title_fontsize)
            if xlab is not None:
                ax.set_xlabel(xlab, fontsize=label_fontsize)
            if ylab is not None:
                ax.set_ylabel(ylab, fontsize=label_fontsize)
        else:
            ValueError("Separator must be either a valid year <int> or 'cids' <str>.")

        if isinstance(ax, plt.Axes):
            ax.tick_params(axis="both", labelsize=tick_fontsize)
        plt.tight_layout()
        if show_plot:
            plt.show()

    def ols_table(self, type="pool"):
        """
        Print statsmodels regression summaries.

        Parameters
        ----------
        type : str
            type of linear regression summary to print. Default is 'pool'. Alternative
            is 're' for period-specific random effects.
        """

        assert type in ["pool", "re"], "Type must be either 'pool' or 're'."

        x, y = self.df.dropna().iloc[:, 0], self.df.dropna().iloc[:, 1]
        x_fit = sm.add_constant(x)
        groups = self.df.reset_index().real_date
        if type == "pool":
            fit_results = sm.OLS(y, x_fit).fit()
        elif type == "re":
            fit_results = sm.MixedLM(y, x_fit, groups).fit(reml=False)

        print(fit_results.summary())


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "NZD", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2000-01-01", "2020-12-31", 0.1, 1]
    df_cids.loc["CAD"] = ["2001-01-01", "2020-11-30", 0, 1]
    df_cids.loc["BRL"] = ["2001-01-01", "2020-11-30", -0.1, 2]
    df_cids.loc["GBP"] = ["2002-01-01", "2020-11-30", 0, 2]
    df_cids.loc["NZD"] = ["2002-01-01", "2020-09-30", -0.1, 2]
    df_cids.loc["USD"] = ["2003-01-01", "2020-12-31", -0.1, 2]

    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]
    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-10-30", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2001-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2001-01-01", "2020-10-30", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {"AUD": ["2000-01-01", "2003-12-31"], "GBP": ["2018-01-01", "2100-01-01"]}

    # All AUD GROWTH locations.
    filt1 = (dfd["xcat"] == "GROWTH") & (dfd["cid"] == "AUD")
    filt2 = (dfd["xcat"] == "INFL") & (dfd["cid"] == "NZD")

    # Reduced DataFrame.
    dfdx = dfd[~(filt1 | filt2)].copy()
    dfdx["ERA"] = "before 2007"
    dfdx.loc[dfdx["real_date"].dt.year > 2007, "ERA"] = "from 2010"

    def modify_cry_values(group):
        if group.name[1] == "CRY":  # Check if xcat is "cry"
            mask = np.ones(len(group), dtype=bool)
            mask[np.arange(len(group)) % 20 != 0] = False  # Keep only every 20th row
            group.loc[~mask, "value"] = 0  # Set all other rows to zero
        return group

    dfdx = dfdx.groupby(["cid", "xcat"], group_keys=False).apply(modify_cry_values)

    cidx = ["AUD", "CAD", "GBP", "USD", "PRY"]

    cr = CategoryRelations(
        dfdx,
        xcats=["CRY", "XR"],
        freq="D",
        lag=1,
        cids=cidx,
        xcat_aggs=["mean", "sum"],
        start="2001-01-01",
        blacklist=black,
        years=None,
    )

    cr.reg_scatter(
        labels=False,
        separator=None,
        title="Carry and Return",
        xlab="Carry",
        ylab="Return",
        coef_box="lower left",
        prob_est="map",
        remove_zero_predictor=True,
    )

    # years parameter

    cr = CategoryRelations(
        dfdx,
        xcats=["CRY", "XR"],
        freq="M",
        years=5,
        lag=0,
        cids=cidx,
        xcat_aggs=["mean", "sum"],
        start="2001-01-01",
        blacklist=black,
    )

    cr.reg_scatter(
        labels=False,
        separator=None,
        title="Carry and Return, 5-year periods",
        xlab="Carry",
        ylab="Return",
        coef_box="lower left",
        prob_est="map",
    )

    cr = CategoryRelations(
        dfdx,
        xcats=["CRY", "XR"],
        # xcat1_chg="diff",
        freq="M",
        lag=1,
        cids=cidx,
        xcat_aggs=["mean", "sum"],
        start="2001-01-01",
        blacklist=black,
        years=None,
    )

    cr.reg_scatter(
        labels=False,
        separator=2010,
        title="Carry and Return",
        xlab="Carry",
        ylab="Return",
        coef_box="lower left",
        ncol=5,
        remove_zero_predictor=True
    )
    cr.reg_scatter(
        labels=False,
        separator="cids",
        title="Composite macro trend pressure indicator and subsequent IRS fixed receiver returns for USD and EUR, since 2000",
        xlab="Composite macro trend pressure indicator",
        ylab="Next month's return on 2-year IRS return, vol-targeted position, %",
        coef_box="lower left",
        ncol=2,
    )

    # Passing Axes object for a subplot
    fig, ax = plt.subplots(1, 2, figsize=(12, 8))

    for i in range(2):
        cr.reg_scatter(
            labels=False,
            separator=None,
            title="Carry and Return",
            xlab="Carry",
            ylab="Return",
            coef_box="lower left",
            prob_est="kendall",
            ax=ax[i],
        )
    plt.show()

    cr.ols_table(type="pool")
    cr.ols_table(type="re")



"""
Module for calculating contract signals based on cross-section-specific signals, and
hedging them with a basket of contracts. Main function is `contract_signals`.
"""

import pandas as pd
import warnings
from numbers import Number
from typing import List, Union, Tuple, Optional, Set, Any, Dict

from macrosynergy.management.types import NoneType, QuantamentalDataFrame
from macrosynergy.panel import make_relative_value
from macrosynergy.management.utils import (
    is_valid_iso_date,
    update_df,
    reduce_df,
    estimate_release_frequency,
)
import logging

logger = logging.getLogger(__name__)


def _check_scaling_args(
    ctypes: List[str],
    cscales: Optional[List[Union[Number, str]]] = None,
    csigns: Optional[List[int]] = None,
    hbasket: Optional[List[str]] = None,
    hscales: Optional[List[Union[Number, str]]] = None,
    hratios: Optional[str] = None,
) -> Tuple[Any, Any, Any, Any, Any]:

    ## Check cscales and csigns
    if cscales is not None:
        # check that the number of scales is the same as the number of ctypes
        if len(cscales) != len(ctypes):
            raise ValueError("`cscales` must be of the same length as `ctypes`")
        if not all([isinstance(x, (str, Number)) for x in cscales]):
            raise TypeError("`cscales` must be a List of strings or numerical values")
    else:
        cscales: List[Number] = [1.0] * len(ctypes)

    if csigns is not None:
        # check that the number of signs is the same as the number of ctypes
        if len(csigns) != len(ctypes):
            raise ValueError("`csigns` must be of the same length as `ctypes`")
        if not all([isinstance(x, int) for x in csigns]):
            raise TypeError("`csigns` must be a List of integers")
        if not all([x in [-1, 1] for x in csigns]):
            warnings.warn(
                "`csigns` must be a List of -1 or 1, coercing to -1 or 1",
                UserWarning,
            )
            csigns: List[int] = [x / abs(x) for x in csigns]
    else:
        csigns: List[int] = [1] * len(ctypes)

    ## Check hbasket and hscales
    if hbasket is not None:
        if not (bool(hratios) and bool(hscales)):
            raise ValueError(
                "`hratios` and `hscales` must be provided if `hbasket` is provided"
            )
        else:
            # check that the number of scales is the same as the number of hbasket
            if len(hscales) != len(hbasket):
                raise ValueError("`hscales` must be of the same length as `hbasket`")
            if not all([isinstance(x, (str, Number)) for x in hscales]):
                raise TypeError(
                    "`hscales` must be a List of strings or numerical values"
                )

    return cscales, csigns, hbasket, hscales, hratios


def _check_estimation_frequency(df_wide: pd.DataFrame, rebal_freq: str) -> pd.DataFrame:
    """
    Check the timeseries to see if the estimated frequency matches the actual frequency.
    Raises a warning if the estimated frequency does not match the rebalancing frequency.

    Parameters
    ----------
    df_wide : pd.DataFrame
        dataframe in wide format with the contract signals.
    est_freq : str
        the estimated frequency of the contract signals.

    Returns
    -------
    pd.DataFrame
        dataframe with the estimated frequency.
    """

    estimated_freq: pd.Series = estimate_release_frequency(df_wide=df_wide)

    # for each series in the dataframe, check if the estimated frequency matches the rebal_freq
    for _col in df_wide.columns:
        if estimated_freq[_col] is None:
            warnings.warn(
                f"Unable to estimate frequency for `{_col}`",
            )
        elif estimated_freq[_col] != rebal_freq:
            warnings.warn(
                f"Estimated frequency for `{_col}` does not match "
                f"the rebalancing frequency`{rebal_freq}`",
            )

    return df_wide


def _make_relative_value(
    df: pd.DataFrame,
    sig: str,
    *args,
    **kwargs,
) -> pd.DataFrame:
    assert isinstance(df, QuantamentalDataFrame), "`df` must be a QuantamentalDataFrame"
    return make_relative_value(
        df=df[df["xcat"] == sig], xcats=[sig], postfix="", *args, **kwargs
    )


def _gen_contract_signals(
    df_wide: pd.DataFrame,
    cids: List[str],
    sig: str,
    ctypes: List[str],
    cscales: List[Union[Number, str]],
    csigns: List[int],
) -> pd.DataFrame:
    """
    Generate contract signals from cross-section-specific signals.

    Parameters
    ----------
    df : pd.DataFrame
        dataframe in quantamental format with the contract signals and potentially
        categories required for translation into contract signals.
    cids : List[str]
        list of cross-sections whose signals are to be translated into contract signals.
    sig : str
        the category ticker of the cross-section-specific signal that is translated into
        contract signals.
    ctypes : List[str]
        list of identifiers for the contract types that are to be traded. They typically
        correspond to the contract type acronyms that are used in JPMaQS for generic
        returns, carry and volatility, such as "FX" for FX forwards or "EQ" for equity index
        futures. N.B. Overall a contract is identified by the combination of its cross-
        section and its contract type "<cid>_<ctype>".
    cscales : List[Union[Number, str]]
        list of scaling factors for the contract signals. These can be either a list of
        floats or a list of category tickers that serve as basis of translation. The former
        are fixed across time, the latter variable.
    csigns : List[int]
        list of signs for the contract signals. These must be either 1 for long position
        or -1 for short position.

    Returns
    -------
    pd.DataFrame
        dataframe with scaling applied.
    """

    expected_contract_signals: List[str] = [f"{cx}_{sig}" for cx in cids]

    # Check that all the contract signals are in the dataframe
    if not set(expected_contract_signals).issubset(set(df_wide.columns)):
        raise ValueError(
            "Some `cids` are missing the `sig` in the provided dataframe."
            f"\nMissing: {set(expected_contract_signals) - set(df_wide.columns)}"
        )

    # Multiply each cid_ctype by the corresponding scale and sign to produce the contract signal
    new_conts: List[str] = []
    for _cid in cids:
        sig_col: str = _cid + "_" + sig
        for ix, ctx in enumerate(ctypes):
            scale_var: Union[Number, pd.Series]
            # If the scale is a string, it must be a category ticker
            # Otherwise it is a fixed numeric value
            if isinstance(cscales[ix], str):
                scale_var: pd.Series = df_wide[_cid + "_" + cscales[ix]]
            else:
                scale_var: Number = cscales[ix]

            new_cont_name: str = _cid + "_" + ctx + "_CSIG"
            df_wide[new_cont_name] = df_wide[sig_col] * csigns[ix] * scale_var
            # For convenience, keep track of the new contract signals
            new_conts.append(new_cont_name)

    # Only return the new contract signals
    df_wide = df_wide.loc[:, new_conts]

    return df_wide


def _apply_hedge_ratios(
    df_wide: pd.DataFrame,
    cids: List[str],
    sig: str,
    hbasket: List[str],
    hscales: List[Union[Number, str]],
    hratios: str,
) -> pd.DataFrame:

    # check if the CID_SIG is in the dataframe
    expc_cid_sigs: List[str] = [f"{cx}_{sig}" for cx in cids]
    expc_cid_hr: List[str] = [f"{cx}_{hratios}" for cx in cids]
    err_str: str = (
        "Some `cids` are missing the `{sig_type}` in the provided dataframe."
        "\nMissing: {missing_items}"
    )
    for sig_type, expc_sigs in zip(["sig", "hratio"], [expc_cid_sigs, expc_cid_hr]):
        if not set(expc_sigs).issubset(set(df_wide.columns)):
            raise ValueError(
                err_str.format(
                    sig_type=sig_type,
                    missing_items=set(expc_sigs) - set(df_wide.columns),
                )
            )
    hedged_assets_list: List[str] = []
    for hb_ix, _hb in enumerate(hbasket):
        basket_pos: str = _hb + "_CSIG"
        df_wide[basket_pos] = 0.0  # initialise the basket position to zero
        for _cid in cids:
            # HBASKETx_CSIG = CIDx_SIG * CIDx_HRATIO * HBASKETx_SCALE
            # e.g.:
            # USD_EQ_CSIG = AUD_SIG * AUD_HRATIO * USD_EQ_HSCALE

            cid_sig: str = _cid + "_" + sig
            cid_hr: str = _cid + "_" + hratios

            hb_hratio: Union[Number, pd.Series]
            # If the scale is a string, it must be a category ticker
            # Otherwise it is a fixed numeric value
            if isinstance(hscales[hb_ix], str):
                hb_hratio: pd.Series = df_wide[_cid + "_" + hscales[hb_ix]]
            else:
                hb_hratio: Number = hscales[hb_ix]

            # Add the basket position to the exisitng basket_pos column in the df
            _posx: pd.Series = df_wide[cid_sig] * df_wide[cid_hr] * hb_hratio
            # set nans to zero, to avoid nans in the final result
            _posx = _posx.fillna(0.0)
            df_wide[basket_pos] += _posx

            hedged_assets_list.append(basket_pos)

    # List(Set(hedged_assets_list)) to remove duplicates
    hedged_assets_list: List[str] = list(set(hedged_assets_list))

    df_wide = df_wide.loc[:, hedged_assets_list]

    return df_wide


def _add_hedged_signals(
    df_wide_cs: pd.DataFrame,
    df_wide_hs: Optional[pd.DataFrame] = None,
) -> pd.DataFrame:
    if df_wide_hs is None:
        return df_wide_cs

    for _col in set(df_wide_hs.columns):
        if _col not in df_wide_cs.columns:
            df_wide_cs[_col] = 0.0
        df_wide_cs[_col] += df_wide_hs[_col]

    return df_wide_cs


def contract_signals(
    df: pd.DataFrame,
    sig: str,
    cids: List[str],
    ctypes: List[str],
    cscales: Optional[List[Union[Number, str]]] = None,
    csigns: Optional[List[int]] = None,
    hbasket: Optional[List[str]] = None,
    hscales: Optional[List[Union[Number, str]]] = None,
    hratios: Optional[str] = None,
    relative_value: bool = False,
    start: Optional[str] = None,
    end: Optional[str] = None,
    rebal_freq: str = "M",
    blacklist: Optional[dict] = None,
    sname: str = "STRAT",
    *args,
    **kwargs,
) -> QuantamentalDataFrame:
    """
    Calculate contract-specific signals based on cross section-specific signals.
    Each contract signal a product of the cross-section-specific signal, a scaling factor
    or category, and a sign (+-1). Optionally, the contract signals can be hedged with a
    basket of contracts. The hedge ratios are determined by a cross-section-specific
    category. When hedging, the hedged signals are added to the contract signals; thus
    yielding a single contract signal per financial contract.

    Parameters
    ----------
    df : pd.DataFrame
        standardized JPMaQS DataFrame with the necessary columns: 'cid', 'xcat',
        'real_date' and 'value'. This dataframe must contain the cross-section-specific
        signals and possibly
            [1] categories for changing scale factors of the main contracts,
            [2] contracts of a hedging basket, and
            [3] cross-section specific hedge ratios
    sig : str
        the cross section-specific signal that serves as the basis of contract signals.
    cids : List[str]
        list of cross-sections whose signal is to be used.
    ctypes : List[str]
        list of identifiers for the contract types that are to be traded. They typically
        correspond to the contract type acronyms that are used in JPMaQS for generic
        returns, carry and volatility, such as "FX" for FX forwards or "EQ" for equity index
        futures. If n contracts are traded per cross sections `ctypes` must contain n
        arguments. N.B. Overall a contract is identified by the combination of its cross-
        section and its contract type "<cid>_<ctype>".
    cscales : List[str|float]
        list of scaling factors for the contract signals. These can be either a list of
        floats or a list of category tickers that serve as basis of translation. The former
        are fixed across time, the latter variable. The list `cscales` must be of the same
        length as the list `ctypes`.
    csigns : List[int]
        list of signs that determine the direction of the contract signals. Contract
        signal is sign x cross section signal. The signs must be either 1 or -1. The list
        `csigns` must be of the same length as `ctypes` and `cscales`.
    hbasket : List[str]
        list of contract identifiers in the format "<cid>_<ctype>" that serve as
        constituents of a hedging basket, if one is used.    
    hscales : List[str|float]
        list of scaling factors (weights) for the basket. These can be either a list of
        floats or a list of category tickers that serve as basis of translation. The former
        are fixed across time, the latter variable.
    hratios : str
        category name for cross-section-specific hedge ratios. The values of this
        category determine direction and size of the hedge basket per unit of the cross
        section-specific signal.
    relative_value : bool
        If False (default), no relative value is calculated. If True boolean, relative
        value is calculated for all cids in the strategy. # TODO split above
        `relative_value` argument into two: `relative_value` and `relative_value_cids`?
    start : str
        earliest date in ISO format. Default is None and earliest date in df is used.
    end : str
        latest date in ISO format. Default is None and latest date in df is used.
    blacklist : dict
        cross-sections with date ranges that should be excluded from the calculation of
        contract signals.
    sname : str
        name of the strategy. Default is "STRAT".

    Returns
    -------
    pd.DataFrame
        with the contract signals for all traded contracts and the specified strategy.
        It has the standard JPMaQS DataFrame. The contract signals have the following format
        "<cid>_<ctype>_<sname>_CSIG".
    """

    ## basic type and value checks
    for varx, namex, typex in [
        (sig, "sig", str),
        (cids, "cids", list),
        (ctypes, "ctypes", list),
        (cscales, "cscales", (list, NoneType)),
        (csigns, "csigns", (list, NoneType)),
        (hbasket, "hbasket", (list, NoneType)),
        (hscales, "hscales", (list, NoneType)),
        (hratios, "hratios", (str, NoneType)),
        (relative_value, "relative_value", bool),
        (start, "start", (str, NoneType)),
        (end, "end", (str, NoneType)),
        (blacklist, "blacklist", (dict, NoneType)),
        (sname, "sname", str),
    ]:
        if not isinstance(varx, typex):
            raise TypeError(f"`{namex}` must be <{typex}> not <{type(varx)}>")

        if typex in [list, str, dict] and len(varx) == 0:
            raise ValueError(f"`{namex}` must not be an empty {str(typex)}")

    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError("`df` must be a standardised quantamental dataframe")

    ## Standardise and copy the dataframe
    df: pd.DataFrame = QuantamentalDataFrame(df)
    _initialized_as_categorical: bool = df.InitializedAsCategorical

    ## Check the dates
    if start is None:
        start: str = pd.Timestamp(df["real_date"].min()).strftime("%Y-%m-%d")
    if end is None:
        end: str = pd.Timestamp(df["real_date"].max()).strftime("%Y-%m-%d")

    for dx, nx in [(start, "start"), (end, "end")]:
        if not is_valid_iso_date(dx):
            raise ValueError(f"`{nx}` must be a valid ISO-8601 date string")

    ## Reduce the dataframe
    df: pd.DataFrame = reduce_df(df=df, start=start, end=end, blacklist=blacklist)

    ## Check that all cid_ctype are in the dataframe
    expected_base_signals: Set[str] = set([f"{cx}_{sig}" for cx in cids])
    found_base_signals: Set[str] = set(QuantamentalDataFrame(df).list_tickers())
    if not (expected_base_signals).issubset(found_base_signals):
        raise ValueError(
            "Some `cids` are missing the `sig` in the provided dataframe."
            f"\nMissing: {expected_base_signals - found_base_signals}"
        )

    ## Check the scaling and hedging arguments
    cscales, csigns, hbasket, hscales, hratios = _check_scaling_args(
        ctypes=ctypes,
        cscales=cscales,
        csigns=csigns,
        hbasket=hbasket,
        hscales=hscales,
        hratios=hratios,
    )

    # Actual calculation

    ## Calculate relative value if requested
    if relative_value:
        rel_signals = _make_relative_value(
            df=df,
            sig=sig,
            blacklist=blacklist,
            cids=cids,
            start=start,
            end=end,
            *args,
            **kwargs,
        )
        df = update_df(df=df, df_add=rel_signals)

    ## Cast the dataframe to wide format
    df_wide: pd.DataFrame = QuantamentalDataFrame(df=df).to_wide()

    ## Check rebal_freq or downsample the dataframe
    # df_wide: pd.DataFrame = _check_estimation_frequency(
    #     df_wide=df_wide, rebal_freq=rebal_freq
    # )

    ## Generate primary contract signals
    df_contract_signals: pd.DataFrame = _gen_contract_signals(
        df_wide=df_wide,
        cids=cids,
        sig=sig,
        ctypes=ctypes,
        cscales=cscales,
        csigns=csigns,
    )

    ## Generate hedge contract signals
    df_hedge_signals: Optional[pd.DataFrame] = None
    if hbasket is not None:
        df_hedge_signals: pd.DataFrame = _apply_hedge_ratios(
            df_wide=df_wide,
            cids=cids,
            sig=sig,
            hbasket=hbasket,
            hscales=hscales,
            hratios=hratios,
        )

    # Add the hedge signals to the contract signals
    df_out: pd.DataFrame = _add_hedged_signals(
        df_wide_cs=df_contract_signals,
        df_wide_hs=df_hedge_signals,
    )

    ## Wide to quantamental
    df_out: pd.DataFrame = QuantamentalDataFrame.from_wide(df=df_out)

    # Append the strategy name to all the xcats
    df_out = df_out.rename_xcats(postfix=f"_{sname}")

    assert isinstance(df_out, QuantamentalDataFrame)
    return QuantamentalDataFrame(
        df_out, _initialized_as_categorical=_initialized_as_categorical
    ).to_original_dtypes()


def multi_signal_contract_signals(
    df: QuantamentalDataFrame,
    contract_dicts: Dict[str, Dict[str, Any]],
    common_args: Dict[str, Any],
):
    """
    Calculate contract signals for multiple signal specifications. Each signal
    specification is a dictionary with arguments for a call to `contract_signals()`. The
    common arguments are passed to all calls. The dataframe is passed to all calls, and
    should not be included in the dictionaries, or common arguments. Please see the
    documentation for `contract_signals()` for more information on the arguments. Note:
    when two (or more) of the arguments produce a series with the same name, each new
    series will overwrite the previous one.

    Parameters
    ----------
    df : pd.DataFrame
        standardized JPMaQS DataFrame with the necessary columns: 'cid', 'xcat',
        'real_date' and 'value'.
    contract_dicts : dict[str, dict[str, Any]]
        dictionary of dictionaries with arguments for each call to `contract_signals()`.
        The keys are the names of the signal specifications.
    common_args : dict[str, Any]
        dictionary with common arguments for all calls to `contract_signals()`. These
        are passed to all calls, and should not be included in the dictionaries in
        `contract_dicts`.

    Returns
    -------
    QuantamentalDataFrame
        dataframe with the contract signals for all traded contracts and the specified
        strategies. It has the standard JPMaQS DataFrame. The contract signals have the
        following format "<cid>_<ctype>_<sname>_CSIG".
    """

    # Check that the types are correct
    types_dict: dict = {
        "sig": str,
        "cids": list,
        "ctypes": list,
        "cscales": (list, NoneType),
        "csigns": (list, NoneType),
        "hbasket": (list, NoneType),
        "hscales": (list, NoneType),
        "hratios": (str, NoneType),
        "relative_value": bool,
        "start": (str, NoneType),
        "end": (str, NoneType),
        "blacklist": (dict, NoneType),
        "sname": str,
    }

    df = QuantamentalDataFrame(df)
    _initialized_as_categorical: bool = df.InitializedAsCategorical

    type_error_msg: str = "{} must be <{}> not <{}>"
    value_error_msg: str = "{} must not be an empty {}"
    for k, v in common_args.items():
        if not isinstance(v, types_dict[k]):
            raise TypeError(type_error_msg.format(k, types_dict[k], type(v)))
        if types_dict[k] in [list, str, dict] and len(v) == 0:
            raise ValueError(value_error_msg.format(k, str(types_dict[k])))

    type_error_msg = type_error_msg + " ; in arguments passed for {}"
    value_error_msg = value_error_msg + " ; in arguments passed for {}"

    for dict_name, dict_args in contract_dicts.items():
        for k, v in dict_args.items():
            if not isinstance(v, types_dict[k]):
                raise TypeError(
                    type_error_msg.format(k, types_dict[k], type(v), dict_name)
                )
            if types_dict[k] in [list, str, dict] and len(v) == 0:
                raise ValueError(
                    value_error_msg.format(k, str(types_dict[k]), dict_name)
                )

    # Calculate the contract signals for each signal specification
    # Keep updating the dataframe with the new signals to allow signals to be used to
    # generate other signals
    # the awkward syntax **{**a, **b} is used to update left dict values with right dict values
    # to avoid conflicts in the keys of the two dictionaries

    for dict_args in contract_dicts.values():
        # TODO: use update_df?
        df = QuantamentalDataFrame.from_qdf_list(
            [df, contract_signals(df=df, **{**common_args, **dict_args})]
        )

    return QuantamentalDataFrame(
        df, _initialized_as_categorical=_initialized_as_categorical
    ).to_original_dtypes()


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_test_df

    cids: List[str] = ["USD", "EUR", "GBP", "AUD", "CAD"]
    xcats: List[str] = ["SIG", "HR"]

    start: str = "2000-01-01"
    end: str = "2020-12-31"

    df: pd.DataFrame = make_test_df(
        cids=cids,
        xcats=xcats,
        start=start,
        end=end,
    )

    df.loc[(df["cid"] == "USD") & (df["xcat"] == "SIG"), "value"] = 1.0
    ctypes = ["FX", "IRS", "CDS"]
    cscales = [1.0, 0.5, 0.1]
    csigns = [1, -1, 1]

    hbasket = ["USD_EQ", "EUR_EQ"]
    hscales = [0.7, 0.3]

    df_cs: pd.DataFrame = contract_signals(
        df=df,
        sig="SIG",
        cids=cids,
        ctypes=ctypes,
        cscales=cscales,
        csigns=csigns,
        hbasket=hbasket,
        hscales=hscales,
        hratios="HR",
    )
    df_cs["xcat"].unique()



from typing import Union
import numpy as np
import pandas as pd
import scipy.stats as stats
from sklearn.base import BaseEstimator
from statsmodels.tools.tools import add_constant

import macrosynergy.management as msm
from macrosynergy.management import make_qdf


class RandomEffects(BaseEstimator):
    """
    Random effects model for inference on panel data.

    Parameters
    ----------
    group_col : str
        The column of a multi-indexed pandas dataframe to group by, for placing of
        random effects.
    fit_intercept : bool
        Whether or not to fit an intercept term.

    Notes
    -----
    A random effects model is a way of attributing variance in a dependent variable to
    variance in a group of observations (i.e. error variance), in a structured manner.
    In the context of panel data, we often use this to account for cross-sectional
    correlations by imposing period-specific effects. However, this model is solely
    for inference and is not predictive.
    """

    def __init__(self, group_col="real_date", fit_intercept=True):
        if not isinstance(group_col, str):
            raise TypeError("group_col must be a string.")
        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")

        self.fit_intercept = fit_intercept
        self.group_col = group_col

        self.params = None
        self.std_errors = None
        self.coef_ = None
        self.intercept_ = None
        self.fitted = None
        self.effects = None
        self.idiosyncratic = None
        self.cov = None
        self.r2 = None
        self.sigma2_residuals = None
        self.sigma2_effects = None
        self.theta = None
        self.residuals = None
        self.residual_ss = None
        self.total_ss = None
        self.nobs = None

    def fit(self, X, y):
        """
        Fit the random effects model.

        Parameters
        ----------
        X : Pandas DataFrame or Series
            Input feature matrix.
        y : Pandas DataFrame or Series
            Target values with the same index as X.

        Returns
        -------
        self
            The fitted estimator.
        """

        X, y = self.check_X_y(X, y)

        if self.fit_intercept:
            X = add_constant(X)

        self._fit(X, y)

        return self

    def _fit(self, df_x, df_y):
        """
        Private method for fitting the random effects model.

        Parameters
        ----------
        df_x : Pandas DataFrame
            Features.
        df_y : Pandas DataFrame
            Targets.
        """

        y_demeaned = self._demean(df_y)
        x_demeaned = self._demean(df_x)

        # Fixed Effect Estimation
        params, _, _, _ = np.linalg.lstsq(x_demeaned, y_demeaned, rcond=None)
        eps = y_demeaned.to_numpy() - x_demeaned.to_numpy() @ params

        # Between Estimation
        xbar = df_x.groupby(level=self.group_col).mean()
        ybar = df_y.groupby(level=self.group_col).mean()
        params, ssr, _, _ = np.linalg.lstsq(xbar, ybar, rcond=None)
        alpha = np.asarray(ybar.values) - np.asarray(xbar.values) @ params

        # Estimate variances
        nobs = float(df_y.shape[0])
        neffects = alpha.shape[0]
        nvars = df_x.shape[1]

        # Idiosyncratic variance
        sigma2_e = float(np.squeeze(eps.T @ eps)) / (nobs - nvars - neffects + 1)

        obs_counts = np.asarray(df_y.groupby(level=self.group_col).count())
        obs_harmonic_mean = neffects / ((1.0 / obs_counts)).sum()

        # Random effect variance
        sigma2_a = max(0.0, (ssr / (neffects - nvars)) - sigma2_e / obs_harmonic_mean)

        # Theta
        theta = 1.0 - np.sqrt(sigma2_e / (obs_counts * sigma2_a + sigma2_e))

        index = df_y.index
        reindex = index.levels[1][index.codes[1]]

        # Random effects estimation
        ybar = (theta * ybar).loc[reindex]
        xbar = (theta * xbar).loc[reindex]

        y = np.asarray(df_y)
        x = np.asarray(df_x)

        y = y - ybar.values
        x = x - xbar.values

        params, residual_ss, _, _ = np.linalg.lstsq(x, y, rcond=None)
        eps = y - x @ params

        # Covariance estimation
        cov_matrix = self._cov(x, residual_ss, nobs)

        index = df_y.index
        fitted = pd.DataFrame(df_x.values @ params, index, ["fitted_values"])
        effects = pd.DataFrame(
            np.asarray(df_y) - np.asarray(fitted) - eps,
            index,
            ["estimated_effects"],
        )
        idiosyncratic = pd.DataFrame(eps, index, ["idiosyncratic"])

        if self.fit_intercept:
            y = y - y.mean(0)

        total_ss = float(np.squeeze(y.T @ y))
        r2 = 1 - residual_ss / total_ss

        self.features = df_x.columns
        self.params = pd.Series(params.reshape(-1), index=self.features, name="params")
        self.std_errors = pd.Series(
            np.sqrt(np.diag(cov_matrix)), index=self.features, name="std_error"
        )

        # For sklearn compatibility
        if self.fit_intercept:
            self.intercept_ = self.params["const"]
            self.coef_ = self.params.drop("const").values.reshape(-1)
        else:
            self.intercept_ = 0
            self.coef_ = self.params.values.reshape(-1)
        self.fitted = fitted
        self.effects = effects
        self.idiosyncratic = idiosyncratic
        self.cov = cov_matrix
        self.r2 = r2
        self.sigma2_residuals = sigma2_e
        self.sigma2_effects = sigma2_a
        self.theta = theta
        self.residuals = eps
        self.residual_ss = residual_ss
        self.total_ss = total_ss
        self.nobs = nobs

    def _demean(self, df):
        """
        Demean the groups as specified by self.group_col.

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe with `self.group_col` as a level in its multiIndex.

        Returns
        -------
        pd.DataFrame
            The demeaned DataFrame.

        Notes
        -----
        If the model is fit with an intercept, the grand mean is added back to the
        demeaned DataFrame.
        """
        mu = df.groupby(level=self.group_col).transform("mean")
        df_demeaned = df - mu
        if self.fit_intercept:
            return df_demeaned + df.mean(0)
        else:
            return df_demeaned

    def _cov(self, x, ssr, nobs):
        """
        Compute the covariance matrix of the parameter estimates.

        Parameters
        ----------
        x : np.ndarray
            The design matrix.
        ssr : float
            The sum of squared residuals.
        nobs : int
            The number of observations.

        Returns
        -------
        np.ndarray
            The estimated covariance matrix.
        """
        s2 = float(ssr.item()) / nobs
        cov = s2 * np.linalg.inv(x.T @ x)  # TODO: Could use pinv?
        return cov

    def check_X_y(self, X, y):
        """
        Checks for the input data.

        Parameters
        ----------
        X : pd.DataFrame or pd.Series
            Input feature matrix
        y : pd.DataFrame or pd.Series
            Target values with the same index as X.

        Returns
        -------
        pd.DataFrame, pd.DataFrame
            The input data.
        """

        X = self._df_checks(X)
        y = self._df_checks(y)

        if not (X.index == y.index).all():
            raise ValueError("Index of X and y must be the same.")

        return X, y

    def _df_checks(self, df):
        """
        Perform checks on the DataFrame input.

        Parameters
        ----------
        df : pd.DataFrame or pd.Series
            The DataFrame to check.

        Returns
        -------
        pd.DataFrame
            The DataFrame with checks performed.
        """
        if isinstance(df, pd.Series):
            df = df.to_frame()

        if not isinstance(df, pd.DataFrame):
            raise TypeError("Input must be a pandas DataFrame.")

        if not isinstance(df.index, pd.MultiIndex):
            raise ValueError("DataFrame must have a MultiIndex.")

        if self.group_col not in df.index.names:
            raise ValueError(f"Group column '{self.group_col}' not found in index.")

        return df

    @property
    def pvals(self):
        """
        Compute the p-values for the parameter estimates.

        Returns
        -------
        pd.Series
            The p-values.
        """
        zstat = self.params / self.std_errors
        pvals = 2 * (1 - stats.norm.cdf(np.abs(zstat)))
        return pd.Series(pvals, index=self.features, name="pvals")


if __name__ == "__main__":

    np.random.seed(1)

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids2 = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids2.loc["AUD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids2.loc["CAD"] = ["2013-01-01", "2020-12-31", 0, 1]
    df_cids2.loc["GBP"] = ["2010-01-01", "2020-12-31", 0, 1]
    df_cids2.loc["USD"] = ["2010-01-01", "2020-12-31", 0, 1]

    df_xcats2 = pd.DataFrame(index=xcats, columns=cols)
    df_xcats2.loc["XR"] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats2.loc["CRY"] = ["2010-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats2.loc["GROWTH"] = ["2010-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats2.loc["INFL"] = ["2010-01-01", "2020-12-31", 1, 2, 0.8, 0.5]

    dfd2 = make_qdf(df_cids2, df_xcats2, back_ar=0.75)
    dfd2["grading"] = np.ones(dfd2.shape[0])
    black = {"GBP": ["2009-01-01", "2012-06-30"], "CAD": ["2018-01-01", "2100-01-01"]}
    dfd2 = msm.reduce_df(df=dfd2, cids=cids, xcats=xcats, blacklist=black)

    dfd2 = dfd2.pivot(index=["cid", "real_date"], columns="xcat", values="value")
    XX = dfd2.drop(columns=["XR"])
    yy = dfd2["XR"]

    X, y = XX.copy(), yy.copy()

    ftrs = []
    feature_names_in_ = np.array(X.columns)

    # Keep significant features
    for col in feature_names_in_[:1]:
        ftr = X[col].copy()

        rem = RandomEffects(group_col="real_date", fit_intercept=True)
        rem.fit(ftr, y)

        print(rem.params)
        print(rem.pvals)



import datetime
import pandas as pd
import warnings
import requests
from typing import List
from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.management.utils import get_cid, get_xcat
from io import StringIO

MAX_RETRIES = 3
CSV_URL = None

def _request_wrapper(url: str, verbose: bool = True, **kwargs) -> str:
    """
    Wrapper around `requests.get` to handle errors and retries.
    Any additional keyword arguments are passed to `requests.get`. Purpose is to allow
    users to pass additional headers, proxy settings, cert verification, etc.
    """

    if verbose:
        print(f"Requesting data from {url}")

    for _ in range(MAX_RETRIES):
        try:
            r = requests.get(url, **kwargs)
            r.raise_for_status()
            return r.text
        except requests.exceptions.RequestException as e:
            if verbose:
                print(f"Error downloading {url}: {e}")
            continue
    raise requests.exceptions.RequestException(f"Failed to download {url}")


AVAILABLE_CTYPES: List[str] = ["FX", "IRS", "CDS"]
AVAIALBLE_COSTS: List[str] = ["BIDOFFER", "ROLLCOST", "SIZE"]
AVAILABLE_STATS: List[str] = ["MEDIAN", "90PCTL"]
AVAILABLE_CATS: List[str] = [
    f"{c}{t}_{s}"
    for c in AVAILABLE_CTYPES
    for t in AVAIALBLE_COSTS
    for s in AVAILABLE_STATS
]


def download_transaction_costs(
    csv_url: str = CSV_URL,
    verbose: bool = False,
    categorical: bool = False,
    **kwargs,
) -> QuantamentalDataFrame:
    """
    Download trading costs data from the S3 bucket.

    Parameters
    ----------
    csv_url : str
        URL of the CSV file to download.
    verbose : bool
        Print progress information.
    **kwargs : args
        Additional keyword arguments to pass to `requests.get`. This can be used to pass
        additional headers, proxy settings, cert verification, etc.
    """
    if csv_url is None or not isinstance(csv_url, str):
        raise ValueError(f"Invalid CSV URL provided: {csv_url}")
    if verbose:
        print(f"Timestamp (UTC): {datetime.datetime.now(datetime.timezone.utc)}")
        print(f"Downloading trading costs data from {csv_url}")
    dfd: pd.DataFrame = pd.read_csv(
        StringIO(_request_wrapper(csv_url, **kwargs)),
        parse_dates=["real_date"],
    )
    dfd["cid"], dfd["xcat"] = get_cid(dfd["ticker"]), get_xcat(dfd["ticker"])
    dfd = dfd[QuantamentalDataFrame.IndexCols + ["value"]]
    if not isinstance(dfd, pd.DataFrame) or dfd.empty:
        raise ValueError("Failed to download trading costs data")

    if not isinstance(dfd, QuantamentalDataFrame):
        if verbose:
            warnings.warn(
                "Downloaded data could not be converted to QuantamentalDataFrame",
                UserWarning,
            )
    else:
        if verbose:
            print("Downloaded data and transformed data successfully")
        dfd = QuantamentalDataFrame(dfd, categorical=categorical)

    return dfd


if __name__ == "__main__":
    print(download_transaction_costs(verbose=True).head())



import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import Tuple, List
from macrosynergy.management import business_day_dif


def view_panel_dates(
    df: pd.DataFrame,
    size: Tuple[float, float] = None,
    use_last_businessday: bool = True,
    header: str = None,
    row_order: List[str] = None,
):
    """
    Visualize panel dates with color codes.

    Parameters
    ----------
    df : ~pandas.DataFrame
        A standardized Quantamental DataFrame with dates as index and series as columns.
    size : Tuple[float, float]
        tuple of floats with width/length of displayed heatmap.
    use_last_businessday : bool
        boolean indicating whether or not to use the last business day before today as
        the end date. Default is True.
    header : str
        A string to be used as the title of the heatmap. If None, a default header will be used
        based on the data type of the DataFrame.
    row_order : List[str]
        A list of strings specifying the order of rows in the heatmap. These rows
        correspond to the columns of the input DataFrame. If None, the default order
        used by Seaborn will be applied.
    """

    # DataFrame of official timestamps.
    if all(df.dtypes == object):
        df = df.apply(pd.to_datetime)
        # All series, in principle, should be populated to the last active release date
        # in the DataFrame.

        if use_last_businessday:
            maxdate: pd.Timestamp = (
                pd.Timestamp.today() - pd.tseries.offsets.BusinessDay()
            )
        else:
            maxdate: pd.Timestamp = df.max().max()

        df = business_day_dif(df=df, maxdate=maxdate)

        df = df.astype(float)
        # Ideally the data type should be int, but Pandas cannot represent NaN as int.
        # -- https://pandas.pydata.org/pandas-docs/stable/user_guide/gotchas.html#support-for-integer-na
        if header is None:
            header = f"Missing days up to {maxdate.strftime('%Y-%m-%d')}"

    else:
        if header is None:
            header = "Start years of quantamental indicators."

    if size is None:
        size = (max(df.shape[0] / 2, 18), max(1, df.shape[1] / 2))

    df = df.T

    if row_order is None:
        row_order = df.index.tolist()

    if isinstance(df.index, pd.CategoricalIndex):
        missing = set(row_order) - set(df.index.categories)
        if missing:
            df = df.reindex(row_order, fill_value=pd.NA)
        df.index = pd.CategoricalIndex(df.index, categories=row_order, ordered=True)
        df = df.sort_index()
    else:
        missing = set(row_order) - set(df.index)
        if missing:
            df = df.reindex(row_order, fill_value=pd.NA)
        df = df.loc[row_order]

    sns.set(rc={"figure.figsize": size})
    sns.heatmap(
        df,
        cmap="YlOrBr",
        center=df.stack().mean(),
        annot=True,
        fmt=".0f",
        linewidth=1,
        cbar=False,
    )
    plt.xlabel("")
    plt.ylabel("")
    plt.title(header, fontsize=18)
    plt.show()



"""
Functions used to visualize correlations across categories or cross-sections of panels.
"""

import pandas as pd
import numpy as np
from typing import Dict, List, Union, Tuple, Optional, Any
from macrosynergy.management.simulate import make_qdf

import macrosynergy.visuals as msv


def correl_matrix(
    df: pd.DataFrame,
    xcats: Union[str, List[str]] = None,
    cids: List[str] = None,
    xcats_secondary: Optional[Union[str, List[str]]] = None,
    cids_secondary: Optional[List[str]] = None,
    start: str = None,
    end: str = None,
    val: str = "value",
    freq: str = None,
    cluster: bool = False,
    lags: dict = None,
    lags_secondary: Optional[dict] = None,
    title: str = None,
    size: Tuple[float] = (14, 8),
    max_color: float = None,
    show: bool = True,
    xcat_labels: Optional[Union[List[str], Dict[str, str]]] = None,
    xcat_secondary_labels: Optional[Union[List[str], Dict[str, str]]] = None,
    cid_labels: Optional[Union[List[str], Dict[str, str]]] = None,
    cid_secondary_labels: Optional[Union[List[str], Dict[str, str]]] = None,
    **kwargs: Any,
):
    """
    Visualize correlation across categories or cross-sections of panels.

    Parameters
    ----------
    df : ~pandas.Dataframe
        standardized JPMaQS DataFrame with the necessary columns: 'cid', 'xcat',
        'real_date' and at least one column with values of interest.
    xcats : List[str]
        extended categories to be correlated. Default is all in the DataFrame. If `xcats`
        contains only one category the correlation coefficients across cross sections are
        displayed. If `xcats` contains more than one category, the correlation coefficients
        across categories are displayed. Additionally, the order of the `xcats` received will
        be mirrored in the correlation matrix.
    cids : List[str]
        cross sections to be correlated. Default is all in the DataFrame.
    xcats_secondary : List[str]
        an optional second set of extended categories. If `xcats_secondary` is provided,
        correlations will be calculated between the categories in `xcats` and `xcats_secondary`.
    cids_secondary : List[str]
        an optional second list of cross sections. If `cids_secondary` is provided
        correlations will be calculated and visualized between these two sets.
    start : str
        earliest date in ISO format. Default is None and earliest date in df is used.
    end : str
        latest date in ISO format. Default is None and latest date in df is used.
    val : str
        name of column that contains the values of interest. Default is 'value'.
    freq : str
        frequency option. Per default the correlations are calculated based on the
        native frequency of the datetimes in 'real_date', which is business daily. Down-
        sampling options include weekly ('W'), monthly ('M'), or quarterly ('Q') mean.
    cluster : bool
        if True the series in the correlation matrix are reordered by hierarchical
        clustering. Default is False.
    lags : dict
        optional dictionary of lags applied to respective categories. The key will be
        the category and the value is the lag or lags. If a category has multiple lags
        applied, pass in a list of lag values. The lag factor will be appended to the
        category name in the correlation matrix. If `xcats_secondary` is not none, this
        parameter will specify lags for the categories in `xcats`.
    lags_secondary : dict
        optional dictionary of lags applied to the second set of categories if
        `xcats_secondary` is provided.
    title : str
        chart heading. If none is given, a default title is used.
    size : Tuple[float]
        two-element tuple setting width/height of figure. Default is (14, 8).
    max_color : float
        maximum values of positive/negative correlation coefficients for color scale.
        Default is none. If a value is given it applies symmetrically to positive and
        negative values.
    show : bool
        if True the figure will be displayed. Default is True.
    xcat_labels : Optional[Union[List[str], Dict[str, str]]
        optional list or dictionary of labels for the categories specified in `xcats`.
        A list should be in the same order as `xcats`, a dictionary should map from each
        category to its label (e.g. {'XR': 'Excess returns', 'CRY': 'Carry'}).
    xcat_secondary_labels : Optional[Union[List[str], Dict[str, str]]]
        optional list or dictionary of labels for `xcats_secondary`.
    cid_labels : Optional[Union[List[str], Dict[str, str]]]
        optional list or dictionary of labels for cids. A list should be in the same
        order as cids, a dictionary should map from each cid to its label.
    cid_secondary_labels : Optional[Union[List[str], Dict[str, str]]]
        optional list or dictionary of labels for cids_secondary.
    **kwargs : Dict
        Arbitrary keyword arguments that are passed to seaborn.heatmap.


    .. note::
        Lags can include a 0 if the original should also be correlated.

    .. note::
        The function displays the heatmap of a correlation matrix across categories or cross-
        sections (depending on which parameter has received multiple elements).
    """

    msv.view_correlation(
        df=df,
        xcats=xcats,
        cids=cids,
        xcats_secondary=xcats_secondary,
        cids_secondary=cids_secondary,
        start=start,
        end=end,
        val=val,
        freq=freq,
        cluster=cluster,
        lags=lags,
        lags_secondary=lags_secondary,
        title=title,
        size=size,
        max_color=max_color,
        show=show,
        xcat_labels=xcat_labels,
        xcat_secondary_labels=xcat_secondary_labels,
        cid_labels=cid_labels,
        cid_secondary_labels=cid_secondary_labels,
        **kwargs,
    )


if __name__ == "__main__":
    np.random.seed(0)

    # Un-clustered correlation matrices.

    cids = ["AUD", "CAD", "GBP", "USD", "NZD", "EUR"]
    cids_dmsc = ["CHF", "NOK", "SEK"]
    cids_dmec = ["DEM", "ESP", "FRF", "ITL", "NLG"]
    cids += cids_dmec
    cids += cids_dmsc
    xcats = ["XR", "CRY"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )

    df_cids.loc["AUD"] = ["2010-01-01", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD"] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-11-30", -0.2, 0.5]
    df_cids.loc["USD"] = ["2010-01-01", "2020-12-30", -0.2, 0.5]
    df_cids.loc["NZD"] = ["2002-01-01", "2020-09-30", -0.1, 2]
    df_cids.loc["EUR"] = ["2002-01-01", "2020-09-30", -0.2, 2]
    df_cids.loc["DEM"] = ["2003-01-01", "2020-09-30", -0.3, 2]
    df_cids.loc["ESP"] = ["2003-01-01", "2020-09-30", -0.1, 2]
    df_cids.loc["FRF"] = ["2003-01-01", "2020-09-30", -0.2, 2]
    df_cids.loc["ITL"] = ["2004-01-01", "2020-09-30", -0.2, 0.5]
    df_cids.loc["NLG"] = ["2003-01-01", "2020-12-30", -0.1, 0.5]
    df_cids.loc["CHF"] = ["2003-01-01", "2020-12-30", -0.3, 2.5]
    df_cids.loc["NOK"] = ["2010-01-01", "2020-12-30", -0.1, 0.5]
    df_cids.loc["SEK"] = ["2010-01-01", "2020-09-30", -0.1, 0.5]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR",] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY",] = ["2010-01-01", "2020-10-30", 1, 2, 0.95, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    start = "2012-01-01"
    end = "2020-09-30"

    lag_dict = {"XR": [0, 2, 5]}

    # Clustered correlation matrices. Test hierarchical clustering.
    correl_matrix(
        df=dfd,
        xcats=["XR", "CRY"],
        xcats_secondary=None,
        cids=cids,
        cids_secondary=None,
        start=start,
        end=end,
        val="value",
        freq=None,
        cluster=True,
        title="Correlation Matrix",
        size=(14, 8),
        max_color=None,
        lags=None,
        lags_secondary=None,
        annot=True,
        fmt=".2f",
    )
